import asyncio
import json
import pytest

from asynctest import TestCase as AsyncTestCase
from asynctest import mock as async_mock

from ...cache.basic import BasicCache
from ...issuer.base import BaseIssuer, IssuerError
from ...ledger.endpoint_type import EndpointType
from ...ledger.indy import (
    BadLedgerRequestError,
    ClosedPoolError,
    ErrorCode,
    IndyErrorHandler,
    IndyError,
    IndyLedger,
    GENESIS_TRANSACTION_PATH,
    LedgerConfigError,
    LedgerError,
    LedgerTransactionError,
    Role,
    TAA_ACCEPTED_RECORD_TYPE,
)
from ...storage.indy import IndyStorage
from ...storage.record import StorageRecord
from ...wallet.base import DIDInfo


class TestRole(AsyncTestCase):
    async def test_role(self):
        assert Role.get(2) is Role.STEWARD
        assert Role.get(0) is Role.TRUSTEE
        assert Role.get(101) is Role.ENDORSER
        assert Role.get(201) is Role.NETWORK_MONITOR
        assert Role.get(None) is Role.USER
        assert Role.get(-1) is None
        assert Role.get("user") is Role.USER
        assert Role.get("steward") is Role.STEWARD
        assert Role.get("trustee") is Role.TRUSTEE
        assert Role.get("endorser") is Role.ENDORSER
        assert Role.get("network_monitor") is Role.NETWORK_MONITOR
        assert Role.get("ROLE_REMOVE") is None

        assert Role.STEWARD.to_indy_num_str() == "2"
        assert Role.TRUSTEE.to_indy_num_str() == "0"
        assert Role.ENDORSER.to_indy_num_str() == "101"
        assert Role.NETWORK_MONITOR.to_indy_num_str() == "201"
        assert Role.USER.to_indy_num_str() is None
        assert Role.ROLE_REMOVE.to_indy_num_str() == ""

        assert Role.STEWARD.token() == "STEWARD"
        assert Role.TRUSTEE.token() == "TRUSTEE"
        assert Role.ENDORSER.token() == "ENDORSER"
        assert Role.NETWORK_MONITOR.token() == "NETWORK_MONITOR"
        assert Role.USER.token() is None
        assert Role.ROLE_REMOVE.to_indy_num_str() == ""


@pytest.mark.indy
class TestIndyLedger(AsyncTestCase):
    test_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_did_info = DIDInfo(
        test_did, "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx", None
    )
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"

    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("builtins.open")
    async def test_init(self, mock_open, mock_create_config):
        mock_open.return_value = async_mock.MagicMock()

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        ledger = IndyLedger("name", mock_wallet)

        assert ledger.pool_name == "name"
        assert ledger.wallet is mock_wallet

        await ledger.create_pool_config("genesis_transactions")

        mock_open.assert_called_once_with(GENESIS_TRANSACTION_PATH, "w")
        mock_open.return_value.__enter__.return_value.write.assert_called_once_with(
            "genesis_transactions"
        )
        mock_create_config.assert_called_once_with(
            "name", json.dumps({"genesis_txn": GENESIS_TRANSACTION_PATH})
        )
        assert ledger.did_to_nym(ledger.nym_to_did(self.test_did)) == self.test_did

    @async_mock.patch("indy.pool.list_pools")
    @async_mock.patch("builtins.open")
    async def test_init_do_not_recreate(self, mock_open, mock_list_pools):
        mock_open.return_value = async_mock.MagicMock()
        mock_list_pools.return_value = [{"pool": "name"}, {"pool": "another"}]

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        ledger = IndyLedger("name", mock_wallet)
        assert ledger.type == "indy"

        assert ledger.pool_name == "name"
        assert ledger.wallet is mock_wallet

        with self.assertRaises(LedgerConfigError):
            await ledger.create_pool_config("genesis_transactions", recreate=False)

        mock_open.assert_called_once_with(GENESIS_TRANSACTION_PATH, "w")

    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.delete_pool_ledger_config")
    @async_mock.patch("indy.pool.list_pools")
    @async_mock.patch("builtins.open")
    async def test_init_recreate(
        self, mock_open, mock_list_pools, mock_delete_config, mock_create_config
    ):
        mock_open.return_value = async_mock.MagicMock()
        mock_list_pools.return_value = [{"pool": "name"}, {"pool": "another"}]
        mock_delete_config.return_value = None

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        ledger = IndyLedger("name", mock_wallet)

        assert ledger.pool_name == "name"
        assert ledger.wallet is mock_wallet

        await ledger.create_pool_config("genesis_transactions", recreate=True)

        mock_open.assert_called_once_with(GENESIS_TRANSACTION_PATH, "w")
        mock_delete_config.assert_called_once_with("name")
        mock_create_config.assert_called_once_with(
            "name", json.dumps({"genesis_txn": GENESIS_TRANSACTION_PATH})
        )

    async def test_init_non_indy(self):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "non-indy"
        with self.assertRaises(LedgerConfigError):
            IndyLedger("name", mock_wallet)

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    async def test_aenter_aexit(
        self, mock_close_pool, mock_open_ledger, mock_set_proto
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        ledger = IndyLedger("name", mock_wallet)

        async with ledger as led:
            mock_set_proto.assert_called_once_with(2)
            mock_open_ledger.assert_called_once_with("name", "{}")
            assert led == ledger
            mock_close_pool.assert_not_called()
            assert led.pool_handle == mock_open_ledger.return_value

        mock_close_pool.assert_called_once()
        assert ledger.pool_handle == None

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    async def test_aenter_aexit_nested_keepalive(
        self, mock_close_pool, mock_open_ledger, mock_set_proto
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        ledger = IndyLedger("name", mock_wallet, keepalive=1)

        async with ledger as led0:
            mock_set_proto.assert_called_once_with(2)
            mock_open_ledger.assert_called_once_with("name", "{}")
            assert led0 == ledger
            mock_close_pool.assert_not_called()
            assert led0.pool_handle == mock_open_ledger.return_value

        async with ledger as led1:
            assert ledger.ref_count == 1

        mock_close_pool.assert_not_called()  # it's a future
        assert ledger.pool_handle

        await asyncio.sleep(1.01)
        mock_close_pool.assert_called_once()
        assert ledger.pool_handle == None

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    async def test_aenter_aexit_close_x(
        self, mock_close_pool, mock_open_ledger, mock_set_proto
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_close_pool.side_effect = IndyError(ErrorCode.PoolLedgerTimeout)
        ledger = IndyLedger("name", mock_wallet)

        with self.assertRaises(LedgerError):
            async with ledger as led:
                assert led.pool_handle == mock_open_ledger.return_value

        assert ledger.pool_handle == mock_open_ledger.return_value
        assert ledger.ref_count == 1

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    async def test_submit_pool_closed(
        self, mock_close_pool, mock_open_ledger, mock_create_config, mock_set_proto
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        ledger = IndyLedger("name", mock_wallet)

        with self.assertRaises(ClosedPoolError) as context:
            await ledger._submit("{}")
        assert "sign and submit request to closed pool" in str(context.exception)

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("indy.ledger.sign_and_submit_request")
    async def test_submit_signed(
        self,
        mock_sign_submit,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_sign_submit.return_value = '{"op": "REPLY"}'

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = None

            with self.assertRaises(BadLedgerRequestError):
                await ledger._submit("{}", True)

            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_did = mock_wallet.get_public_did.return_value
            mock_did.did = self.test_did

            await ledger._submit("{}", True, False)

            mock_wallet.get_public_did.assert_called_once_with()

            mock_sign_submit.assert_called_once_with(
                ledger.pool_handle, mock_wallet.handle, mock_did.did, "{}"
            )

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("indy.ledger.sign_and_submit_request")
    @async_mock.patch("indy.ledger.append_txn_author_agreement_acceptance_to_request")
    async def test_submit_signed_taa_accept(
        self,
        mock_append_taa,
        mock_sign_submit,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_append_taa.return_value = "{}"
        mock_sign_submit.return_value = '{"op": "REPLY"}'

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        ledger = IndyLedger("name", mock_wallet)
        ledger.get_latest_txn_author_acceptance = async_mock.CoroutineMock(
            return_value={
                "text": "sample",
                "version": "0.0",
                "digest": "digest",
                "mechanism": "dummy",
                "time": "now",
            }
        )

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_did = mock_wallet.get_public_did.return_value
            mock_did.did = self.test_did

            await ledger._submit(
                request_json="{}",
                sign=None,
                taa_accept=True,
                sign_did=self.test_did_info,
            )

            mock_wallet.get_public_did.assert_not_called()
            mock_append_taa.assert_called_once_with(
                "{}", "sample", "0.0", "digest", "dummy", "now"
            )
            mock_sign_submit.assert_called_once_with(
                ledger.pool_handle, ledger.wallet.handle, self.test_did, "{}"
            )

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("indy.ledger.submit_request")
    async def test_submit_unsigned(
        self,
        mock_submit,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_did = async_mock.MagicMock()

        future = asyncio.Future()
        future.set_result(mock_did)

        mock_submit.return_value = '{"op": "REPLY"}'

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did.return_value = future

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            await ledger._submit("{}", False)

            mock_wallet.get_public_did.assert_not_called()

            mock_submit.assert_called_once_with(ledger.pool_handle, "{}")

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("indy.ledger.submit_request")
    async def test_submit_unsigned_ledger_transaction_error(
        self,
        mock_submit,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_did = async_mock.MagicMock()

        future = asyncio.Future()
        future.set_result(mock_did)

        mock_submit.return_value = '{"op": "NO-SUCH-OP"}'

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did.return_value = future

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with self.assertRaises(LedgerTransactionError):
                await ledger._submit("{}", False)

            mock_wallet.get_public_did.assert_not_called()

            mock_submit.assert_called_once_with(ledger.pool_handle, "{}")

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("indy.ledger.submit_request")
    async def test_submit_rejected(
        self,
        mock_submit,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_did = async_mock.MagicMock()

        future = asyncio.Future()
        future.set_result(mock_did)

        mock_submit.return_value = '{"op": "REQNACK", "reason": "a reason"}'

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did.return_value = future

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with self.assertRaises(LedgerTransactionError) as context:
                await ledger._submit("{}", False)
            assert "Ledger rejected transaction request" in str(context.exception)

        mock_submit.return_value = '{"op": "REJECT", "reason": "another reason"}'

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did.return_value = future

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with self.assertRaises(LedgerTransactionError) as context:
                await ledger._submit("{}", False)
            assert "Ledger rejected transaction request" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.fetch_schema_by_id")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.fetch_schema_by_seq_no")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_schema_request")
    async def test_send_schema(
        self,
        mock_build_schema_req,
        mock_add_record,
        mock_fetch_schema_by_seq_no,
        mock_fetch_schema_by_id,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        issuer = async_mock.MagicMock(BaseIssuer)
        ledger = IndyLedger("name", mock_wallet)

        issuer.create_schema.return_value = ("schema_issuer_did:name:1.0", "{}")
        mock_fetch_schema_by_id.return_value = None
        mock_fetch_schema_by_seq_no.return_value = None

        mock_submit.return_value = (
            r'{"op":"REPLY","result":{"txnMetadata":{"seqNo": 1}}}'
        )

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = None

            with self.assertRaises(BadLedgerRequestError):
                schema_id, schema_def = await ledger.create_and_send_schema(
                    issuer, "schema_name", "schema_version", [1, 2, 3]
                )

            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_did = mock_wallet.get_public_did.return_value
            mock_did.did = self.test_did

            schema_id, schema_def = await ledger.create_and_send_schema(
                issuer, "schema_name", "schema_version", [1, 2, 3]
            )

            mock_wallet.get_public_did.assert_called_once_with()
            issuer.create_schema.assert_called_once_with(
                mock_did.did, "schema_name", "schema_version", [1, 2, 3]
            )

            mock_build_schema_req.assert_called_once_with(
                mock_did.did, issuer.create_schema.return_value[1]
            )

            mock_submit.assert_called_once_with(
                mock_build_schema_req.return_value,
                True,
                sign_did=mock_wallet.get_public_did.return_value,
            )

            assert schema_id == issuer.create_schema.return_value[0]

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.check_existing_schema")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_schema_request")
    async def test_send_schema_already_exists(
        self,
        mock_build_schema_req,
        mock_add_record,
        mock_check_existing,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):
        # mock_did = async_mock.CoroutineMock()

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_wallet.get_public_did.return_value.did = "abc"

        fetch_schema_id = (
            f"{mock_wallet.get_public_did.return_value.did}:2:"
            "schema_name:schema_version"
        )
        mock_check_existing.return_value = (fetch_schema_id, {})

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.create_schema.return_value = ("1", "{}")
        ledger = IndyLedger("name", mock_wallet)

        with async_mock.patch.object(
            ledger, "get_indy_storage", async_mock.MagicMock()
        ) as mock_get_storage:
            mock_add_record = async_mock.CoroutineMock()
            mock_get_storage.return_value = async_mock.MagicMock(
                add_record=mock_add_record
            )

            async with ledger:
                schema_id, schema_def = await ledger.create_and_send_schema(
                    issuer, "schema_name", "schema_version", [1, 2, 3]
                )
                assert schema_id == fetch_schema_id
                assert schema_def == {}

            mock_add_record.assert_not_called()

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.check_existing_schema")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_schema_request")
    async def test_send_schema_ledger_transaction_error_already_exists(
        self,
        mock_build_schema_req,
        mock_add_record,
        mock_check_existing,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_wallet.get_public_did.return_value.did = "abc"

        fetch_schema_id = (
            f"{mock_wallet.get_public_did.return_value.did}:2:"
            "schema_name:schema_version"
        )
        mock_check_existing.side_effect = [None, (fetch_schema_id, "{}")]

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.create_schema.return_value = ("1", "{}")
        ledger = IndyLedger("name", mock_wallet)
        ledger._submit = async_mock.CoroutineMock(
            side_effect=LedgerTransactionError("UnauthorizedClientRequest")
        )

        async with ledger:
            schema_id, schema_def = await ledger.create_and_send_schema(
                issuer, "schema_name", "schema_version", [1, 2, 3]
            )
            assert schema_id == fetch_schema_id

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.check_existing_schema")
    async def test_send_schema_ledger_read_only(
        self,
        mock_check_existing,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_wallet.get_public_did.return_value.did = "abc"

        fetch_schema_id = (
            f"{mock_wallet.get_public_did.return_value.did}:2:"
            "schema_name:schema_version"
        )
        mock_check_existing.side_effect = [None, fetch_schema_id]

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.create_schema.return_value = ("1", "{}")
        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            with self.assertRaises(LedgerError) as context:
                await ledger.create_and_send_schema(
                    issuer, "schema_name", "schema_version", [1, 2, 3]
                )
            assert "read only" in str(context.exception)

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.check_existing_schema")
    async def test_send_schema_issuer_error(
        self,
        mock_check_existing,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_wallet.get_public_did.return_value.did = "abc"

        fetch_schema_id = (
            f"{mock_wallet.get_public_did.return_value.did}:2:"
            "schema_name:schema_version"
        )
        mock_check_existing.side_effect = [None, fetch_schema_id]

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.create_schema = async_mock.CoroutineMock(
            side_effect=IssuerError("dummy error")
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with self.assertRaises(LedgerError) as context:
                await ledger.create_and_send_schema(
                    issuer, "schema_name", "schema_version", [1, 2, 3]
                )
            assert "dummy error" in str(context.exception)

    @async_mock.patch("indy.pool.set_protocol_version")
    @async_mock.patch("indy.pool.create_pool_ledger_config")
    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.check_existing_schema")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_schema_request")
    async def test_send_schema_ledger_transaction_error(
        self,
        mock_build_schema_req,
        mock_add_record,
        mock_check_existing,
        mock_close_pool,
        mock_open_ledger,
        mock_create_config,
        mock_set_proto,
    ):

        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_wallet.get_public_did.return_value.did = "abc"

        fetch_schema_id = (
            f"{mock_wallet.get_public_did.return_value.did}:2:"
            "schema_name:schema_version"
        )
        mock_check_existing.side_effect = [None, fetch_schema_id]

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.create_schema.return_value = ("1", "{}")
        ledger = IndyLedger("name", mock_wallet)
        ledger._submit = async_mock.CoroutineMock(
            side_effect=LedgerTransactionError("Some other error message")
        )

        async with ledger:
            with self.assertRaises(LedgerTransactionError):
                await ledger.create_and_send_schema(
                    issuer, "schema_name", "schema_version", [1, 2, 3]
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.fetch_schema_by_id")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.fetch_schema_by_seq_no")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_schema_request")
    async def test_send_schema_no_seq_no(
        self,
        mock_build_schema_req,
        mock_add_record,
        mock_fetch_schema_by_seq_no,
        mock_fetch_schema_by_id,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        issuer = async_mock.MagicMock(BaseIssuer)
        ledger = IndyLedger("name", mock_wallet)

        issuer.create_schema.return_value = ("schema_issuer_did:name:1.0", "{}")
        mock_fetch_schema_by_id.return_value = None
        mock_fetch_schema_by_seq_no.return_value = None

        mock_submit.return_value = (
            r'{"op":"REPLY","result":{"txnMetadata":{"no": "seqNo"}}}'
        )

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_did = mock_wallet.get_public_did.return_value
            mock_did.did = self.test_did

            with self.assertRaises(LedgerError) as context:
                await ledger.create_and_send_schema(
                    issuer, "schema_name", "schema_version", [1, 2, 3]
                )
            assert "schema sequence number" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.fetch_schema_by_id")
    async def test_check_existing_schema(
        self,
        mock_fetch_schema_by_id,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value
        mock_did.did = self.test_did

        mock_fetch_schema_by_id.return_value = {"attrNames": ["a", "b", "c"]}

        ledger = IndyLedger("name", mock_wallet)
        async with ledger:
            schema_id, schema_def = await ledger.check_existing_schema(
                public_did=self.test_did,
                schema_name="test",
                schema_version="1.0",
                attribute_names=["c", "b", "a"],
            )
            assert schema_id == f"{self.test_did}:2:test:1.0"

            with self.assertRaises(LedgerTransactionError):
                await ledger.check_existing_schema(
                    public_did=self.test_did,
                    schema_name="test",
                    schema_version="1.0",
                    attribute_names=["a", "b", "c", "d"],
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_schema_request")
    @async_mock.patch("indy.ledger.parse_get_schema_response")
    async def test_get_schema(
        self,
        mock_parse_get_schema_resp,
        mock_build_get_schema_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value
        mock_did.did = self.test_did

        mock_parse_get_schema_resp.return_value = (None, '{"attrNames": ["a", "b"]}')

        mock_submit.return_value = '{"result":{"seqNo":1}}'

        ledger = IndyLedger("name", mock_wallet, cache=BasicCache())

        async with ledger:
            response = await ledger.get_schema("schema_id")

            mock_wallet.get_public_did.assert_called_once_with()
            mock_build_get_schema_req.assert_called_once_with(mock_did.did, "schema_id")
            mock_submit.assert_called_once_with(
                mock_build_get_schema_req.return_value, sign_did=mock_did
            )
            mock_parse_get_schema_resp.assert_called_once_with(mock_submit.return_value)

            assert response == json.loads(mock_parse_get_schema_resp.return_value[1])

            response == await ledger.get_schema("schema_id")  # cover get-from-cache
            assert response == json.loads(mock_parse_get_schema_resp.return_value[1])

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_schema_request")
    async def test_get_schema_not_found(
        self,
        mock_build_get_schema_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value
        mock_did.did = self.test_did

        mock_submit.return_value = json.dumps({"result": {"seqNo": None}})

        ledger = IndyLedger("name", mock_wallet, cache=BasicCache())

        async with ledger:
            response = await ledger.get_schema("schema_id")

            mock_wallet.get_public_did.assert_called_once_with()
            mock_build_get_schema_req.assert_called_once_with(mock_did.did, "schema_id")
            mock_submit.assert_called_once_with(
                mock_build_get_schema_req.return_value, sign_did=mock_did
            )

            assert response is None

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_txn_request")
    @async_mock.patch("indy.ledger.build_get_schema_request")
    @async_mock.patch("indy.ledger.parse_get_schema_response")
    async def test_get_schema_by_seq_no(
        self,
        mock_parse_get_schema_resp,
        mock_build_get_schema_req,
        mock_build_get_txn_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value
        mock_did.did = self.test_did

        mock_parse_get_schema_resp.return_value = (None, '{"attrNames": ["a", "b"]}')

        submissions = [
            json.dumps(
                {
                    "result": {
                        "data": {
                            "txn": {
                                "type": "101",
                                "metadata": {"from": self.test_did},
                                "data": {
                                    "data": {"name": "preferences", "version": "1.0"}
                                },
                            }
                        }
                    }
                }
            ),
            json.dumps({"result": {"seqNo": 999}}),
        ]  # need to subscript these in assertions later
        mock_submit.side_effect = [
            sub for sub in submissions
        ]  # becomes list iterator, unsubscriptable, in mock object

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            response = await ledger.get_schema("999")

            mock_wallet.get_public_did.assert_called_once_with()
            mock_build_get_txn_req.assert_called_once_with(None, None, seq_no=999)
            mock_build_get_schema_req.assert_called_once_with(
                mock_did.did, f"{self.test_did}:2:preferences:1.0"
            )
            mock_submit.assert_has_calls(
                [
                    async_mock.call(mock_build_get_txn_req.return_value),
                    async_mock.call(
                        mock_build_get_schema_req.return_value, sign_did=mock_did
                    ),
                ]
            )
            mock_parse_get_schema_resp.assert_called_once_with(submissions[1])

            assert response == json.loads(mock_parse_get_schema_resp.return_value[1])

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_txn_request")
    @async_mock.patch("indy.ledger.build_get_schema_request")
    @async_mock.patch("indy.ledger.parse_get_schema_response")
    async def test_get_schema_by_wrong_seq_no(
        self,
        mock_parse_get_schema_resp,
        mock_build_get_schema_req,
        mock_build_get_txn_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value
        mock_did.did = self.test_did

        mock_parse_get_schema_resp.return_value = (None, '{"attrNames": ["a", "b"]}')

        submissions = [
            json.dumps(
                {
                    "result": {
                        "data": {
                            "txn": {
                                "type": "102",
                            }
                        }
                    }
                }
            ),  # not a schema
            json.dumps({"result": {"seqNo": 999}}),
        ]  # need to subscript these in assertions later
        mock_submit.side_effect = [
            sub for sub in submissions
        ]  # becomes list iterator, unsubscriptable, in mock object

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with self.assertRaises(LedgerTransactionError):
                await ledger.get_schema("999")

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_cred_def_request")
    async def test_send_credential_definition(
        self,
        mock_build_cred_def,
        mock_add_record,
        mock_search_records,
        mock_submit,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[]
        )

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.side_effect = [None, cred_def]

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.make_credential_definition_id.return_value = cred_def_id
        issuer.create_and_store_credential_definition.return_value = (
            cred_def_id,
            cred_def_json,
        )
        issuer.credential_definition_in_wallet.return_value = False
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = None

            with self.assertRaises(BadLedgerRequestError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = DIDInfo(
                self.test_did, self.test_verkey, None
            )
            mock_did = mock_wallet.get_public_did.return_value

            (
                result_id,
                result_def,
                novel,
            ) = await ledger.create_and_send_credential_definition(
                issuer, schema_id, None, tag
            )
            assert result_id == cred_def_id
            assert novel

            mock_wallet.get_public_did.assert_called_once_with()
            mock_get_schema.assert_called_once_with(schema_id)

            mock_build_cred_def.assert_called_once_with(mock_did.did, cred_def_json)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_cred_def_request")
    async def test_send_credential_definition_exists_in_ledger_and_wallet(
        self,
        mock_build_cred_def,
        mock_add_record,
        mock_search_records,
        mock_submit,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[]
        )

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = {"mock": "cred-def"}

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.make_credential_definition_id.return_value = cred_def_id
        issuer.create_and_store_credential_definition.return_value = (
            cred_def_id,
            cred_def_json,
        )
        issuer.credential_definition_in_wallet.return_value = True
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        with async_mock.patch.object(
            ledger, "get_indy_storage", async_mock.MagicMock()
        ) as mock_get_storage:
            mock_get_storage.return_value = async_mock.MagicMock(
                add_record=async_mock.CoroutineMock()
            )

            async with ledger:
                mock_wallet.get_public_did = async_mock.CoroutineMock()
                mock_wallet.get_public_did.return_value = DIDInfo(
                    self.test_did, self.test_verkey, None
                )
                mock_did = mock_wallet.get_public_did.return_value

                (
                    result_id,
                    result_def,
                    novel,
                ) = await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )
                assert result_id == cred_def_id
                assert not novel

                mock_wallet.get_public_did.assert_called_once_with()
                mock_get_schema.assert_called_once_with(schema_id)

                mock_build_cred_def.assert_not_called()
                mock_get_storage.assert_not_called()

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    async def test_send_credential_definition_no_such_schema(
        self,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_get_schema.return_value = {}

        issuer = async_mock.MagicMock(BaseIssuer)
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_cred_def_request")
    async def test_send_credential_definition_offer_exception(
        self,
        mock_build_cred_def,
        mock_add_record,
        mock_search_records,
        mock_submit,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[]
        )

        mock_get_schema.return_value = {"seqNo": 999}

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.credential_definition_in_wallet.side_effect = IssuerError(
            "common IO error"
        )
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    async def test_send_credential_definition_cred_def_in_wallet_not_ledger(
        self,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = {}

        issuer = async_mock.MagicMock(BaseIssuer)
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    async def test_send_credential_definition_cred_def_not_on_ledger_wallet_check_x(
        self,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = {}

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.credential_definition_in_wallet = async_mock.CoroutineMock(
            side_effect=IssuerError("dummy error")
        )
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError) as context:
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )
            assert "dummy error" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    async def test_send_credential_definition_cred_def_not_on_ledger_nor_wallet_send_x(
        self,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = {}

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.credential_definition_in_wallet = async_mock.CoroutineMock(
            return_value=False
        )
        issuer.create_and_store_credential_definition = async_mock.CoroutineMock(
            side_effect=IssuerError("dummy error")
        )
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError) as context:
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )
            assert "dummy error" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    async def test_send_credential_definition_read_only(
        self,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = {}

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.credential_definition_in_wallet = async_mock.CoroutineMock(
            return_value=False
        )
        issuer.create_and_store_credential_definition = async_mock.CoroutineMock(
            return_value=("cred-def-id", "cred-def-json")
        )
        ledger = IndyLedger("name", mock_wallet, read_only=True)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError) as context:
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )
            assert "read only" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    async def test_send_credential_definition_cred_def_on_ledger_not_in_wallet(
        self,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = cred_def

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.credential_definition_in_wallet = async_mock.CoroutineMock(
            return_value=False
        )
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()

            with self.assertRaises(LedgerError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_cred_def_request")
    async def test_send_credential_definition_on_ledger_in_wallet(
        self,
        mock_build_cred_def,
        mock_add_record,
        mock_search_records,
        mock_submit,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[]
        )

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = cred_def

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.make_credential_definition_id.return_value = cred_def_id
        issuer.create_and_store_credential_definition.return_value = (
            cred_def_id,
            cred_def_json,
        )
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = None

            with self.assertRaises(BadLedgerRequestError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = DIDInfo(
                self.test_did, self.test_verkey, None
            )
            mock_did = mock_wallet.get_public_did.return_value

            (
                result_id,
                result_def,
                novel,
            ) = await ledger.create_and_send_credential_definition(
                issuer, schema_id, None, tag
            )
            assert result_id == cred_def_id

            mock_wallet.get_public_did.assert_called_once_with()
            mock_get_schema.assert_called_once_with(schema_id)

            mock_build_cred_def.assert_not_called()

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch(
        "aries_cloudagent.ledger.indy.IndyLedger.fetch_credential_definition"
    )
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("indy.ledger.build_cred_def_request")
    async def test_send_credential_definition_create_cred_def_exception(
        self,
        mock_build_cred_def,
        mock_add_record,
        mock_search_records,
        mock_submit,
        mock_fetch_cred_def,
        mock_close,
        mock_open,
        mock_get_schema,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[]
        )

        mock_get_schema.return_value = {"seqNo": 999}
        cred_def_id = f"{self.test_did}:3:CL:999:default"
        cred_def_value = {
            "primary": {"n": "...", "s": "...", "r": "...", "revocation": None}
        }
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": "999",
            "type": "CL",
            "tag": "default",
            "value": cred_def_value,
        }
        cred_def_json = json.dumps(cred_def)

        mock_fetch_cred_def.return_value = None

        issuer = async_mock.MagicMock(BaseIssuer)
        issuer.create_and_store_credential_definition.side_effect = IssuerError(
            "invalid structure"
        )
        ledger = IndyLedger("name", mock_wallet)

        schema_id = "schema_issuer_did:name:1.0"
        tag = "default"

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock()
            mock_wallet.get_public_did.return_value = DIDInfo(
                self.test_did, self.test_verkey, None
            )

            with self.assertRaises(LedgerError):
                await ledger.create_and_send_credential_definition(
                    issuer, schema_id, None, tag
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_cred_def_request")
    @async_mock.patch("indy.ledger.parse_get_cred_def_response")
    async def test_get_credential_definition(
        self,
        mock_parse_get_cred_def_resp,
        mock_build_get_cred_def_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value

        mock_parse_get_cred_def_resp.return_value = (
            None,
            json.dumps({"result": {"seqNo": 1}}),
        )

        ledger = IndyLedger("name", mock_wallet, cache=BasicCache())

        async with ledger:
            response = await ledger.get_credential_definition("cred_def_id")

            mock_wallet.get_public_did.assert_called_once_with()
            mock_build_get_cred_def_req.assert_called_once_with(
                mock_did.did, "cred_def_id"
            )
            mock_submit.assert_called_once_with(
                mock_build_get_cred_def_req.return_value, sign_did=mock_did
            )
            mock_parse_get_cred_def_resp.assert_called_once_with(
                mock_submit.return_value
            )

            assert response == json.loads(mock_parse_get_cred_def_resp.return_value[1])

            response == await ledger.get_credential_definition(  # cover get-from-cache
                "cred_def_id"
            )
            assert response == json.loads(mock_parse_get_cred_def_resp.return_value[1])

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_cred_def_request")
    @async_mock.patch("indy.ledger.parse_get_cred_def_response")
    async def test_get_credential_definition_ledger_not_found(
        self,
        mock_parse_get_cred_def_resp,
        mock_build_get_cred_def_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value

        mock_parse_get_cred_def_resp.side_effect = IndyError(
            error_code=ErrorCode.LedgerNotFound, error_details={"message": "not today"}
        )

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            response = await ledger.get_credential_definition("cred_def_id")

            mock_wallet.get_public_did.assert_called_once_with()
            mock_build_get_cred_def_req.assert_called_once_with(
                mock_did.did, "cred_def_id"
            )
            mock_submit.assert_called_once_with(
                mock_build_get_cred_def_req.return_value, sign_did=mock_did
            )
            mock_parse_get_cred_def_resp.assert_called_once_with(
                mock_submit.return_value
            )

            assert response is None

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_cred_def_request")
    @async_mock.patch("indy.ledger.parse_get_cred_def_response")
    async def test_fetch_credential_definition_ledger_x(
        self,
        mock_parse_get_cred_def_resp,
        mock_build_get_cred_def_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_wallet.get_public_did = async_mock.CoroutineMock()
        mock_did = mock_wallet.get_public_did.return_value

        mock_parse_get_cred_def_resp.side_effect = IndyError(
            error_code=ErrorCode.CommonInvalidParam1,
            error_details={"message": "not today"},
        )

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with self.assertRaises(LedgerError) as context:
                await ledger.fetch_credential_definition("cred_def_id")
            assert "not today" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_key_for_did(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_submit.return_value = json.dumps(
            {"result": {"data": json.dumps({"verkey": self.test_verkey})}}
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_key_for_did(self.test_did)

            assert mock_build_get_nym_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did)
            )
            assert mock_submit.called_once_with(
                mock_build_get_nym_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response == self.test_verkey

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_endpoint_for_did(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = "http://aries.ca"
        mock_submit.return_value = json.dumps(
            {"result": {"data": json.dumps({"endpoint": {"endpoint": endpoint}})}}
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_endpoint_for_did(self.test_did)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response == endpoint

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_endpoint_of_type_profile_for_did(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = "http://company.com/masterdata"
        endpoint_type = EndpointType.PROFILE
        mock_submit.return_value = json.dumps(
            {
                "result": {
                    "data": json.dumps(
                        {"endpoint": {EndpointType.PROFILE.indy: endpoint}}
                    )
                }
            }
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_endpoint_for_did(self.test_did, endpoint_type)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response == endpoint

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_all_endpoints_for_did(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        profile_endpoint = "http://company.com/masterdata"
        default_endpoint = "http://agent.company.com"
        data_json = json.dumps(
            {"endpoint": {"endpoint": default_endpoint, "profile": profile_endpoint}}
        )
        mock_submit.return_value = json.dumps({"result": {"data": data_json}})
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_all_endpoints_for_did(self.test_did)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response == json.loads(data_json).get("endpoint")

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_all_endpoints_for_did_none(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        profile_endpoint = "http://company.com/masterdata"
        default_endpoint = "http://agent.company.com"
        mock_submit.return_value = json.dumps({"result": {"data": None}})
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_all_endpoints_for_did(self.test_did)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response is None

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_endpoint_for_did_address_none(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_submit.return_value = json.dumps(
            {"result": {"data": json.dumps({"endpoint": None})}}
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_endpoint_for_did(self.test_did)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response is None

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_endpoint_for_did_no_endpoint(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_submit.return_value = json.dumps({"result": {"data": None}})
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_endpoint_for_did(self.test_did)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert response is None

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("indy.ledger.build_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_update_endpoint_for_did(
        self,
        mock_submit,
        mock_build_attrib_req,
        mock_build_get_attrib_req,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = ["http://old.aries.ca", "http://new.aries.ca"]
        mock_submit.side_effect = [
            json.dumps(
                {
                    "result": {
                        "data": json.dumps({"endpoint": {"endpoint": endpoint[i]}})
                    }
                }
            )
            for i in range(len(endpoint))
        ]
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.update_endpoint_for_did(self.test_did, endpoint[1])

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            mock_submit.assert_has_calls(
                [
                    async_mock.call(
                        mock_build_get_attrib_req.return_value,
                        sign_did=mock_wallet.get_public_did.return_value,
                    ),
                    async_mock.call(mock_build_attrib_req.return_value, True, True),
                ]
            )
            assert response

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("indy.ledger.build_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_update_endpoint_for_did_no_prior_endpoints(
        self,
        mock_submit,
        mock_build_attrib_req,
        mock_build_get_attrib_req,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = "http://new.aries.ca"
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            with async_mock.patch.object(
                ledger, "get_all_endpoints_for_did", async_mock.CoroutineMock()
            ) as mock_get_all:
                mock_get_all.return_value = None
                mock_wallet.get_public_did = async_mock.CoroutineMock(
                    return_value=self.test_did_info
                )
                response = await ledger.update_endpoint_for_did(self.test_did, endpoint)

                assert mock_build_get_attrib_req.called_once_with(
                    self.test_did,
                    ledger.did_to_nym(self.test_did),
                    "endpoint",
                    None,
                    None,
                )
                mock_submit.assert_has_calls(
                    [
                        async_mock.call(mock_build_attrib_req.return_value, True, True),
                    ]
                )
                assert response

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("indy.ledger.build_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_update_endpoint_of_type_profile_for_did(
        self,
        mock_submit,
        mock_build_attrib_req,
        mock_build_get_attrib_req,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = ["http://company.com/oldProfile", "http://company.com/newProfile"]
        endpoint_type = EndpointType.PROFILE
        mock_submit.side_effect = [
            json.dumps(
                {
                    "result": {
                        "data": json.dumps(
                            {"endpoint": {endpoint_type.indy: endpoint[i]}}
                        )
                    }
                }
            )
            for i in range(len(endpoint))
        ]
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.update_endpoint_for_did(
                self.test_did, endpoint[1], endpoint_type
            )

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            mock_submit.assert_has_calls(
                [
                    async_mock.call(
                        mock_build_get_attrib_req.return_value,
                        sign_did=mock_wallet.get_public_did.return_value,
                    ),
                    async_mock.call(mock_build_attrib_req.return_value, True, True),
                ]
            )
            assert response

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_update_endpoint_for_did_duplicate(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = "http://aries.ca"
        mock_submit.return_value = json.dumps(
            {"result": {"data": json.dumps({"endpoint": {"endpoint": endpoint}})}}
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.update_endpoint_for_did(self.test_did, endpoint)

            assert mock_build_get_attrib_req.called_once_with(
                self.test_did, ledger.did_to_nym(self.test_did), "endpoint", None, None
            )
            assert mock_submit.called_once_with(
                mock_build_get_attrib_req.return_value,
                sign_did=mock_wallet.get_public_did.return_value,
            )
            assert not response

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_attrib_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_update_endpoint_for_did_read_only(
        self, mock_submit, mock_build_get_attrib_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        endpoint = "http://aries.ca"
        mock_submit.return_value = json.dumps(
            {"result": {"data": json.dumps({"endpoint": {"endpoint": endpoint}})}}
        )
        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            with self.assertRaises(LedgerError) as context:
                await ledger.update_endpoint_for_did(self.test_did, "distinct endpoint")
            assert "read only" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_register_nym(
        self, mock_submit, mock_build_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock(
            type="indy",
            get_local_did=async_mock.CoroutineMock(),
            replace_local_did_metadata=async_mock.CoroutineMock(),
        )

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            await ledger.register_nym(self.test_did, self.test_verkey, "alias", None)

            assert mock_build_nym_req.called_once_with(
                self.test_did, self.test_did, self.test_verkey, "alias", None
            )
            assert mock_submit.called_once_with(
                mock_build_nym_req.return_value,
                True,
                True,
                sign_did=mock_wallet.get_public_did.return_value,
            )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_register_nym_ledger_x(
        self, mock_submit, mock_build_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_build_nym_req.side_effect = IndyError(
            error_code=ErrorCode.CommonInvalidParam1,
            error_details={"message": "not today"},
        )

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            with self.assertRaises(LedgerError):
                await ledger.register_nym(
                    self.test_did, self.test_verkey, "alias", None
                )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_register_nym_read_only(
        self, mock_submit, mock_build_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            with self.assertRaises(LedgerError) as context:
                await ledger.register_nym(
                    self.test_did, self.test_verkey, "alias", None
                )
            assert "read only" in str(context.exception)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_nym_role(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_submit.return_value = json.dumps(
            {
                "result": {
                    "dest": "GjZWsBLgZCR18aL468JAT7w9CZRiBnpxUPPgyQxh4voa",
                    "txnTime": 1597858571,
                    "reqId": 1597858571783588400,
                    "state_proof": {
                        "root_hash": "7K26MUQt8E2X1vsRJUmc2298VtY8YC5BSDfT5CRJeUDi",
                        "proof_nodes": "+QHo...",
                        "multi_signature": {
                            "participants": ["Node4", "Node3", "Node2"],
                            "value": {
                                "state_root_hash": "7K2...",
                                "pool_state_root_hash": "GT8...",
                                "ledger_id": 1,
                                "txn_root_hash": "Hnr...",
                                "timestamp": 1597858571,
                            },
                            "signature": "QuX...",
                        },
                    },
                    "data": json.dumps(
                        {
                            "dest": "GjZWsBLgZCR18aL468JAT7w9CZRiBnpxUPPgyQxh4voa",
                            "identifier": "V4SGRU86Z58d6TV7PBUe6f",
                            "role": 101,
                            "seqNo": 11,
                            "txnTime": 1597858571,
                            "verkey": "GjZWsBLgZCR18aL468JAT7w9CZRiBnpxUPPgyQxh4voa",
                        }
                    ),
                    "seqNo": 11,
                    "identifier": "GjZWsBLgZCR18aL468JAT7w9CZRiBnpxUPPgyQxh4voa",
                    "type": "105",
                },
                "op": "REPLY",
            }
        )

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            assert await ledger.get_nym_role(self.test_did) == Role.ENDORSER

            assert mock_build_get_nym_req.called_once_with(self.test_did, self.test_did)
            assert mock_submit.called_once_with(mock_build_get_nym_req.return_value)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    async def test_get_nym_role_indy_x(
        self, mock_build_get_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_build_get_nym_req.side_effect = IndyError(
            error_code=ErrorCode.CommonInvalidParam1,
            error_details={"message": "not today"},
        )
        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )

            with self.assertRaises(LedgerError) as context:
                await ledger.get_nym_role(self.test_did)
            assert "not today" in context.exception.message

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_nym_role_did_not_public_x(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        mock_submit.return_value = json.dumps(
            {
                "result": {
                    "dest": "GjZWsBLgZCR18aL468JAT7w9CZRiBnpxUPPgyQxh4voa",
                    "txnTime": 1597858571,
                    "reqId": 1597858571783588400,
                    "state_proof": {
                        "root_hash": "7K26MUQt8E2X1vsRJUmc2298VtY8YC5BSDfT5CRJeUDi",
                        "proof_nodes": "+QHo...",
                        "multi_signature": {
                            "participants": ["Node4", "Node3", "Node2"],
                            "value": {
                                "state_root_hash": "7K2...",
                                "pool_state_root_hash": "GT8...",
                                "ledger_id": 1,
                                "txn_root_hash": "Hnr...",
                                "timestamp": 1597858571,
                            },
                            "signature": "QuX...",
                        },
                    },
                    "data": json.dumps(None),
                    "seqNo": 11,
                    "identifier": "GjZWsBLgZCR18aL468JAT7w9CZRiBnpxUPPgyQxh4voa",
                    "type": "105",
                },
                "op": "REPLY",
            }
        )

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            with self.assertRaises(BadLedgerRequestError):
                await ledger.get_nym_role(self.test_did)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    @async_mock.patch("indy.ledger.build_get_txn_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.register_nym")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_rotate_public_did_keypair(
        self,
        mock_submit,
        mock_register_nym,
        mock_build_get_txn_request,
        mock_build_get_nym_request,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock(
            get_public_did=async_mock.CoroutineMock(return_value=self.test_did_info),
            rotate_did_keypair_start=async_mock.CoroutineMock(
                return_value=self.test_verkey
            ),
            rotate_did_keypair_apply=async_mock.CoroutineMock(return_value=None),
        )
        mock_wallet.type = "indy"
        mock_submit.side_effect = [
            json.dumps({"result": {"data": json.dumps({"seqNo": 1234})}}),
            json.dumps(
                {
                    "result": {
                        "data": {"txn": {"data": {"role": "101", "alias": "Billy"}}}
                    }
                }
            ),
        ]

        ledger = IndyLedger("name", mock_wallet)
        async with ledger:
            await ledger.rotate_public_did_keypair()

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_rotate_public_did_keypair_no_nym(
        self, mock_submit, mock_build_get_nym_request, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock(
            get_public_did=async_mock.CoroutineMock(return_value=self.test_did_info),
            rotate_did_keypair_start=async_mock.CoroutineMock(
                return_value=self.test_verkey
            ),
            rotate_did_keypair_apply=async_mock.CoroutineMock(return_value=None),
        )
        mock_wallet.type = "indy"
        mock_submit.return_value = json.dumps({"result": {"data": json.dumps(None)}})

        ledger = IndyLedger("name", mock_wallet)
        async with ledger:
            with self.assertRaises(BadLedgerRequestError):
                await ledger.rotate_public_did_keypair()

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_nym_request")
    @async_mock.patch("indy.ledger.build_get_txn_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.register_nym")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_rotate_public_did_keypair_corrupt_nym_txn(
        self,
        mock_submit,
        mock_register_nym,
        mock_build_get_txn_request,
        mock_build_get_nym_request,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock(
            get_public_did=async_mock.CoroutineMock(return_value=self.test_did_info),
            rotate_did_keypair_start=async_mock.CoroutineMock(
                return_value=self.test_verkey
            ),
            rotate_did_keypair_apply=async_mock.CoroutineMock(return_value=None),
        )
        mock_wallet.type = "indy"
        mock_submit.side_effect = [
            json.dumps({"result": {"data": json.dumps({"seqNo": 1234})}}),
            json.dumps({"result": {"data": None}}),
        ]

        ledger = IndyLedger("name", mock_wallet)
        async with ledger:
            with self.assertRaises(BadLedgerRequestError):
                await ledger.rotate_public_did_keypair()

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_revoc_reg_def_request")
    @async_mock.patch("indy.ledger.parse_get_revoc_reg_def_response")
    async def test_get_revoc_reg_def(
        self,
        mock_indy_parse_get_rrdef_resp,
        mock_indy_build_get_rrdef_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_parse_get_rrdef_resp.return_value = ("rr-id", '{"hello": "world"}')

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )

            result = await ledger.get_revoc_reg_def("rr-id")
            assert result == {"hello": "world"}

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_revoc_reg_def_request")
    async def test_get_revoc_reg_def_indy_x(
        self, mock_indy_build_get_rrdef_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_get_rrdef_req.side_effect = IndyError(
            error_code=ErrorCode.CommonInvalidParam1,
            error_details={"message": "not today"},
        )

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )

            with self.assertRaises(IndyError) as context:
                await ledger.get_revoc_reg_def("rr-id")
            assert "not today" in context.exception.message

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_revoc_reg_request")
    @async_mock.patch("indy.ledger.parse_get_revoc_reg_response")
    async def test_get_revoc_reg_entry(
        self,
        mock_indy_parse_get_rr_resp,
        mock_indy_build_get_rr_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_parse_get_rr_resp.return_value = (
            "rr-id",
            '{"hello": "world"}',
            1234567890,
        )

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )

            (result, _) = await ledger.get_revoc_reg_entry("rr-id", 1234567890)
            assert result == {"hello": "world"}

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_get_revoc_reg_delta_request")
    @async_mock.patch("indy.ledger.parse_get_revoc_reg_delta_response")
    async def test_get_revoc_reg_delta(
        self,
        mock_indy_parse_get_rrd_resp,
        mock_indy_build_get_rrd_req,
        mock_submit,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_parse_get_rrd_resp.return_value = (
            "rr-id",
            '{"hello": "world"}',
            1234567890,
        )

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )

            (result, _) = await ledger.get_revoc_reg_delta("rr-id")
            assert result == {"hello": "world"}

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_revoc_reg_def_request")
    async def test_send_revoc_reg_def_public_did(
        self, mock_indy_build_rrdef_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_rrdef_req.return_value = '{"hello": "world"}'

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            await ledger.send_revoc_reg_def({"rr": "def"}, issuer_did=None)
            mock_wallet.get_public_did.assert_called_once()
            assert not mock_wallet.get_local_did.called
            mock_submit.assert_called_once_with(
                mock_indy_build_rrdef_req.return_value, True, True, self.test_did_info
            )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_revoc_reg_def_request")
    async def test_send_revoc_reg_def_local_did(
        self, mock_indy_build_rrdef_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_rrdef_req.return_value = '{"hello": "world"}'

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_local_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            await ledger.send_revoc_reg_def({"rr": "def"}, issuer_did=self.test_did)
            mock_wallet.get_local_did.assert_called_once_with(self.test_did)
            assert not mock_wallet.get_public_did.called
            mock_submit.assert_called_once_with(
                mock_indy_build_rrdef_req.return_value, True, True, self.test_did_info
            )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_revoc_reg_def_request")
    async def test_send_revoc_reg_def_x_no_did(
        self, mock_indy_build_rrdef_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_rrdef_req.return_value = '{"hello": "world"}'

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_local_did = async_mock.CoroutineMock(return_value=None)
            with self.assertRaises(LedgerTransactionError) as context:
                await ledger.send_revoc_reg_def({"rr": "def"}, issuer_did=self.test_did)
            assert "No issuer DID found for revocation registry definition" in str(
                context.exception
            )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_revoc_reg_entry_request")
    async def test_send_revoc_reg_entry_public_did(
        self, mock_indy_build_rre_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_rre_req.return_value = '{"hello": "world"}'

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            await ledger.send_revoc_reg_entry(
                "rr-id", "CL_ACCUM", {"rev-reg": "entry"}, issuer_did=None
            )
            mock_wallet.get_public_did.assert_called_once()
            assert not mock_wallet.get_local_did.called
            mock_submit.assert_called_once_with(
                mock_indy_build_rre_req.return_value, True, True, self.test_did_info
            )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_revoc_reg_entry_request")
    async def test_send_revoc_reg_entry_local_did(
        self, mock_indy_build_rre_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_rre_req.return_value = '{"hello": "world"}'

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_local_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            result = await ledger.send_revoc_reg_entry(
                "rr-id", "CL_ACCUM", {"rev-reg": "entry"}, issuer_did=self.test_did
            )
            mock_wallet.get_local_did.assert_called_once_with(self.test_did)
            assert not mock_wallet.get_public_did.called
            mock_submit.assert_called_once_with(
                mock_indy_build_rre_req.return_value, True, True, self.test_did_info
            )

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    @async_mock.patch("indy.ledger.build_revoc_reg_entry_request")
    async def test_send_revoc_reg_entry_x_no_did(
        self, mock_indy_build_rre_req, mock_submit, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"
        mock_indy_build_rre_req.return_value = '{"hello": "world"}'

        ledger = IndyLedger("name", mock_wallet, read_only=True)

        async with ledger:
            mock_wallet.get_local_did = async_mock.CoroutineMock(return_value=None)
            with self.assertRaises(LedgerTransactionError) as context:
                await ledger.send_revoc_reg_entry(
                    "rr-id", "CL_ACCUM", {"rev-reg": "entry"}, issuer_did=self.test_did
                )
            assert "No issuer DID found for revocation registry entry" in str(
                context.exception
            )

    @async_mock.patch("indy.pool.open_pool_ledger")
    @async_mock.patch("indy.pool.close_pool_ledger")
    async def test_taa_digest_bad_value(
        self,
        mock_close_pool,
        mock_open_ledger,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )

            with self.assertRaises(ValueError):
                await ledger.taa_digest(None, None)

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("indy.ledger.build_get_acceptance_mechanisms_request")
    @async_mock.patch("indy.ledger.build_get_txn_author_agreement_request")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._submit")
    async def test_get_txn_author_agreement(
        self,
        mock_submit,
        mock_build_get_taa_req,
        mock_build_get_acc_mech_req,
        mock_close,
        mock_open,
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        txn_result_data = {"text": "text", "version": "1.0"}
        mock_submit.side_effect = [
            json.dumps({"result": {"data": txn_result_data}}) for i in range(2)
        ]

        ledger = IndyLedger("name", mock_wallet)

        async with ledger:
            mock_wallet.get_public_did = async_mock.CoroutineMock(
                return_value=self.test_did_info
            )
            response = await ledger.get_txn_author_agreement(reload=True)

            assert mock_build_get_acc_mech_req.called_once_with(
                self.test_did, None, None
            )
            assert mock_build_get_taa_req.called_once_with(self.test_did, None)
            mock_submit.assert_has_calls(
                [
                    async_mock.call(
                        mock_build_get_acc_mech_req.return_value,
                        sign_did=mock_wallet.get_public_did.return_value,
                    ),
                    async_mock.call(
                        mock_build_get_taa_req.return_value,
                        sign_did=mock_wallet.get_public_did.return_value,
                    ),
                ]
            )
            assert response == {
                "aml_record": txn_result_data,
                "taa_record": {
                    **txn_result_data,
                    "digest": ledger.taa_digest(
                        txn_result_data["version"], txn_result_data["text"]
                    ),
                },
                "taa_required": True,
            }

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.add_record")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    async def test_accept_and_get_latest_txn_author_agreement(
        self, mock_search_records, mock_add_record, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        ledger = IndyLedger("name", mock_wallet, cache=BasicCache())

        accept_time = ledger.taa_rough_timestamp()
        taa_record = {
            "text": "text",
            "version": "1.0",
            "digest": "abcd1234",
        }
        acceptance = {
            "text": taa_record["text"],
            "version": taa_record["version"],
            "digest": taa_record["digest"],
            "mechanism": "dummy",
            "time": accept_time,
        }

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[
                StorageRecord(
                    TAA_ACCEPTED_RECORD_TYPE,
                    json.dumps(acceptance),
                    {"pool_name": ledger.pool_name},
                )
            ]
        )

        async with ledger:
            await ledger.accept_txn_author_agreement(
                taa_record=taa_record, mechanism="dummy", accept_time=None
            )

            await ledger.cache.clear(f"{TAA_ACCEPTED_RECORD_TYPE}::{ledger.pool_name}")
            for i in range(2):  # populate, then get from, cache
                response = await ledger.get_latest_txn_author_acceptance()
                assert response == acceptance

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.storage.indy.IndyStorage.search_records")
    async def test_get_latest_txn_author_agreement_none(
        self, mock_search_records, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        ledger = IndyLedger("name", mock_wallet, cache=BasicCache())

        mock_search_records.return_value.fetch_all = async_mock.CoroutineMock(
            return_value=[]
        )

        async with ledger:
            await ledger.cache.clear(f"{TAA_ACCEPTED_RECORD_TYPE}::{ledger.pool_name}")
            response = await ledger.get_latest_txn_author_acceptance()
            assert response == {}

    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_open")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger._context_close")
    @async_mock.patch("aries_cloudagent.ledger.indy.IndyLedger.get_schema")
    async def test_credential_definition_id2schema_id(
        self, mock_get_schema, mock_close, mock_open
    ):
        mock_wallet = async_mock.MagicMock()
        mock_wallet.type = "indy"

        S_ID = f"{TestIndyLedger.test_did}:2:favourite_drink:1.0"
        SEQ_NO = "9999"
        mock_get_schema.return_value = {"id": S_ID}

        ledger = IndyLedger("name", mock_wallet, cache=BasicCache())

        async with ledger:
            s_id_short = await ledger.credential_definition_id2schema_id(
                f"{TestIndyLedger.test_did}:3:CL:{SEQ_NO}:tag"
            )

            mock_get_schema.assert_called_once_with(SEQ_NO)

            assert s_id_short == S_ID
            s_id_long = await ledger.credential_definition_id2schema_id(
                f"{TestIndyLedger.test_did}:3:CL:{s_id_short}:tag"
            )
            assert s_id_long == s_id_short

    def test_error_handler(self):
        try:  # with self.assertRaises() makes a copy of exception, loses traceback!
            with IndyErrorHandler("message", LedgerTransactionError):
                try:
                    1 / 0
                except ZeroDivisionError as zx:
                    ix = IndyError(error_code=1, error_details={"message": "bye"})
                    ix.__traceback__ = zx.__traceback__
                    raise ix
        except LedgerTransactionError as err:
            assert type(err) == LedgerTransactionError
            assert type(err.__cause__) == IndyError
            assert err.__traceback__
            assert "bye" in err.message



from asynctest import TestCase as AsyncTestCase
from asynctest import mock as async_mock

from .. import help as command


class TestHelp(AsyncTestCase):
    def test_exec_help(self):
        with async_mock.patch.object(
            command.ArgumentParser, "print_help"
        ) as mock_print_help, async_mock.patch(
            "builtins.print", async_mock.MagicMock()
        ) as mock_print:
            command.execute([])
            mock_print_help.assert_called_once()

            command.execute(["-v"])
            mock_print.assert_called_once_with(command.__version__)

    def test_main(self):
        with async_mock.patch.object(
            command, "__name__", "__main__"
        ) as mock_name, async_mock.patch.object(
            command, "execute", async_mock.MagicMock()
        ) as mock_execute:
            command.main()
            mock_execute.assert_called_once



from asynctest import TestCase as AsyncTestCase
from asynctest import mock as async_mock

from aries_cloudagent.config.injection_context import InjectionContext
from aries_cloudagent.wallet.base import BaseWallet
from aries_cloudagent.wallet.basic import BasicWallet

from ....storage.base import BaseStorage

from .. import routes as test_module


# TODO: Add tests
class TestJSONLDRoutes(AsyncTestCase):
    async def setUp(self):
        self.context = InjectionContext(enforce_typing=False)
        self.storage = async_mock.create_autospec(BaseStorage)
        self.context.injector.bind_instance(BaseStorage, self.storage)
        self.wallet = self.wallet = BasicWallet()
        self.did_info = await self.wallet.create_local_did()
        self.context.injector.bind_instance(BaseWallet, self.wallet)
        self.app = {
            "request_context": self.context,
        }

    async def test_verify_credential(self):
        mock_request = async_mock.MagicMock(
            app=self.app,
            json=async_mock.CoroutineMock(
                return_value={  # posted json
                    "verkey": (
                        # pulled from the did:key in example
                        "5yKdnU7ToTjAoRNDzfuzVTfWBH38qyhE1b9xh4v8JaWF"
                    ),
                    "doc": {
                        "@context": [
                            "https://www.w3.org/2018/credentials/v1",
                            "https://www.w3.org/2018/credentials/examples/v1",
                        ],
                        "id": "http://example.gov/credentials/3732",
                        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
                        "issuer": (
                            "did:key:z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"
                        ),
                        "issuanceDate": "2020-03-10T04:24:12.164Z",
                        "credentialSubject": {
                            "id": (
                                "did:key:"
                                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"
                            ),
                            "degree": {
                                "type": "BachelorDegree",
                                "name": "Bachelor of Science and Arts",
                            },
                        },
                        "proof": {
                            "type": "Ed25519Signature2018",
                            "created": "2020-04-10T21:35:35Z",
                            "verificationMethod": (
                                "did:key:"
                                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc"
                                "4tXLt9DoHd#z6MkjRagNiMu91DduvCvgEsqLZD"
                                "VzrJzFrwahc4tXLt9DoHd"
                            ),
                            "proofPurpose": "assertionMethod",
                            "jws": (
                                "eyJhbGciOiJFZERTQSIsImI2NCI6ZmFsc2UsImNyaX"
                                "QiOlsiYjY0Il19..l9d0YHjcFAH2H4dB9xlWFZQLUp"
                                "ixVCWJk0eOt4CXQe1NXKWZwmhmn9OQp6YxX0a2Lffe"
                                "gtYESTCJEoGVXLqWAA"
                            ),
                        },
                    },
                }
            ),
        )

        with async_mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.verify(mock_request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with({"valid": True})  # expected response

    async def test_sign_credential(self):
        mock_request = async_mock.MagicMock(
            app=self.app,
            json=async_mock.CoroutineMock(
                return_value={  # posted json
                    "verkey": self.did_info.verkey,
                    "doc": {
                        "credential": {
                            "@context": [
                                "https://www.w3.org/2018/credentials/v1",
                                "https://www.w3.org/2018/credentials/examples/v1",
                            ],
                            "id": "http://example.gov/credentials/3732",
                            "type": [
                                "VerifiableCredential",
                                "UniversityDegreeCredential",
                            ],
                            "issuer": (
                                "did:key:"
                                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"
                            ),
                            "issuanceDate": "2020-03-10T04:24:12.164Z",
                            "credentialSubject": {
                                "id": (
                                    "did:key:"
                                    "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"
                                ),
                                "degree": {
                                    "type": "BachelorDegree",
                                    "name": u"Bachelor of Encyclopædic Arts",
                                },
                            },
                        },
                        "options": {
                            "type": "Ed25519Signature2018",
                            "created": "2020-04-10T21:35:35Z",
                            "verificationMethod": (
                                "did:key:"
                                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd#"
                                "z6MkjRagNiMu91DduvCvgEsqLZDVzrJzFrwahc4tXLt9DoHd"
                            ),
                            "proofPurpose": "assertionMethod",
                        },
                    },
                }
            ),
        )

        with async_mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.sign(mock_request)
            assert result == mock_response.return_value
            mock_response.assert_called_once()
            assert "signed_doc" in mock_response.call_args[0][0]
            assert "error" not in mock_response.call_args[0][0]



"""Represents an explicit ack message as per Aries RFC 15."""

from marshmallow import EXCLUDE, fields

from ..agent_message import AgentMessage, AgentMessageSchema


class Ack(AgentMessage):
    """
    Base class representing an explicit ack message.

    Subclass to adopt, specify Meta message type and handler class.
    """

    class Meta:
        """Ack metadata."""

        schema_class = "AckSchema"

    def __init__(self, status: str = None, **kwargs):
        """
        Initialize an explicit ack message instance.

        Args:
            status: Status (default OK)

        """
        super().__init__(**kwargs)
        self.status = status or "OK"


class AckSchema(AgentMessageSchema):
    """Schema for Ack base class."""

    class Meta:
        """Ack schema metadata."""

        model_class = Ack
        unknown = EXCLUDE

    status = fields.Constant(
        constant="OK",
        required=True,
        description="Status: specify OK",
        default="OK",
        example="OK",
    )



"""Indy error handling."""

from typing import Type

from indy.error import IndyError

from ..core.error import BaseError


class IndyErrorHandler:
    """Trap IndyError and raise an appropriate LedgerError instead."""

    def __init__(self, message: str = None, error_cls: Type[BaseError] = BaseError):
        """Init the context manager."""
        self.error_cls = error_cls
        self.message = message

    def __enter__(self):
        """Enter the context manager."""
        return self

    def __exit__(self, err_type, err_value, err_traceback):
        """Exit the context manager."""
        if isinstance(err_value, IndyError):
            raise IndyErrorHandler.wrap_error(
                err_value, self.message, self.error_cls
            ) from err_value

    @classmethod
    def wrap_error(
        cls,
        err_value: IndyError,
        message: str = None,
        error_cls: Type[BaseError] = BaseError,
    ) -> BaseError:
        """Create an instance of BaseError from an IndyError."""
        err_msg = message or "Exception while performing indy operation"
        indy_message = hasattr(err_value, "message") and err_value.message
        if indy_message:
            err_msg += f": {indy_message}"
        err = error_cls(err_msg)
        err.__traceback__ = err_value.__traceback__
        return err



from unittest import TestCase as UnitTestCase

from ......messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema

from ..presentation_exchange import V10PresentationExchange


class BasexRecordImpl(BaseExchangeRecord):
    class Meta:
        schema_class = "BasexRecordImplSchema"

    RECORD_TYPE = "record"
    CACHE_ENABLED = True


class BasexRecordImplSchema(BaseExchangeSchema):
    class Meta:
        model_class = BasexRecordImpl


class TestRecord(UnitTestCase):
    def test_record(self):
        record = V10PresentationExchange(
            presentation_exchange_id="pxid",
            connection_id="connid",
            thread_id="thid",
            initiator="init",
            role="role",
            state="state",
            presentation_proposal_dict={"prop": "dict"},
            presentation_request={"pres": "req"},
            presentation_request_dict={"pres", "dict"},
            presentation={"pres": "indy"},
            verified=False,
            auto_present=True,
            error_msg="error",
        )

        assert record.presentation_exchange_id == "pxid"

        assert record.record_value == {
            "connection_id": "connid",
            "initiator": "init",
            "presentation_proposal_dict": {"prop": "dict"},
            "presentation_request": {"pres": "req"},
            "presentation_request_dict": {"pres", "dict"},
            "presentation": {"pres": "indy"},
            "role": "role",
            "state": "state",
            "auto_present": True,
            "error_msg": "error",
            "verified": False,
            "trace": False,
        }

        bx_record = BasexRecordImpl()
        assert record != bx_record



"""Default ledger provider classes."""

import logging

from ..cache.base import BaseCache
from ..config.base import BaseProvider, BaseInjector, BaseSettings
from ..utils.classloader import ClassLoader
from ..wallet.base import BaseWallet

LOGGER = logging.getLogger(__name__)


class LedgerProvider(BaseProvider):
    """Provider for the default ledger implementation."""

    LEDGER_CLASSES = {"indy": "aries_cloudagent.ledger.indy.IndyLedger"}

    async def provide(self, settings: BaseSettings, injector: BaseInjector):
        """Create and open the ledger instance."""

        pool_name = settings.get("ledger.pool_name", "default")
        keepalive = int(settings.get("ledger.keepalive", 5))
        read_only = bool(settings.get("ledger.read_only", False))
        if read_only:
            LOGGER.error("Note: setting ledger to read-only mode")
        wallet = await injector.inject(BaseWallet)
        ledger = None

        if wallet.type == "indy":
            IndyLedger = ClassLoader.load_class(self.LEDGER_CLASSES["indy"])
            cache = await injector.inject(BaseCache, required=False)
            ledger = IndyLedger(
                pool_name, wallet, keepalive=keepalive, cache=cache, read_only=read_only
            )

            genesis_transactions = settings.get("ledger.genesis_transactions")
            if genesis_transactions:
                await ledger.create_pool_config(genesis_transactions, True)
            elif not await ledger.check_pool_config():
                LOGGER.info("Ledger pool configuration has not been created")
                ledger = None

        return ledger



"""Wallet configuration."""

import logging

import aiohttp

from ..wallet.base import BaseWallet
from ..wallet.crypto import seed_to_did

from .base import ConfigError
from .injection_context import InjectionContext

LOGGER = logging.getLogger(__name__)


async def wallet_config(context: InjectionContext, provision: bool = False):
    """Initialize the wallet."""
    wallet: BaseWallet = await context.inject(BaseWallet)
    if provision:
        if wallet.type != "indy":
            raise ConfigError("Cannot provision a non-Indy wallet type")
        if wallet.created:
            print("Created new wallet")
        else:
            print("Opened existing wallet")
        print("Wallet type:", wallet.type)
        print("Wallet name:", wallet.name)

    wallet_seed = context.settings.get("wallet.seed")
    wallet_local_did = context.settings.get("wallet.local_did")

    public_did_info = None
    public_did = None

    public_did = await wallet.create_local_did()

    genesis_url = context.settings.get("ledger.genesis_url")
    genesis_url = genesis_url.lower()

    if "igrant" in genesis_url:
        if public_did:
            # write it to indy ledger
            async with aiohttp.ClientSession() as session:
                async with session.post('https://indy.igrant.io/register?ca=true', json={"verkey": public_did.verkey, "did": public_did.did, "role": "TRUSTEE" if context.settings.get("ada_mode") else "ENDORSER"}) as resp:
                    if resp.status == 200 or resp.status == 201:
                        print("===============================================================")
                        print("Public DID        : ", public_did.did)
                        print("Public DID verkey : ", public_did.verkey)
                        print("===============================================================")

                        # make it as public
                        public_did_info = await wallet.set_public_did(public_did.did)
    else:
        public_did_info = public_did

    if public_did_info:
        public_did = public_did_info.did
        if wallet_seed and seed_to_did(wallet_seed) != public_did:
            if context.settings.get("wallet.replace_public_did"):
                replace_did_info = await wallet.create_local_did(wallet_seed)
                public_did = replace_did_info.did
                await wallet.set_public_did(public_did)
                print(f"Created new public DID: {public_did}")
                print(f"Verkey: {replace_did_info.verkey}")
            else:
                # If we already have a registered public did and it doesn't match
                # the one derived from `wallet_seed` then we error out.
                raise ConfigError(
                    "New seed provided which doesn't match the registered"
                    + f" public did {public_did}"
                )
        # wait until ledger config to set public DID endpoint - wallet goes first
    elif wallet_seed:
        if wallet_local_did:
            endpoint = context.settings.get("default_endpoint")
            metadata = {"endpoint": endpoint} if endpoint else None

            local_did_info = await wallet.create_local_did(
                seed=wallet_seed, metadata=metadata
            )
            local_did = local_did_info.did
            if provision:
                print(f"Created new local DID: {local_did}")
                print(f"Verkey: {local_did_info.verkey}")
        else:
            public_did_info = await wallet.create_public_did(seed=wallet_seed)
            public_did = public_did_info.did
            if provision:
                print(f"Created new public DID: {public_did}")
                print(f"Verkey: {public_did_info.verkey}")
            # wait until ledger config to set public DID endpoint - wallet goes first

    if provision and not wallet_local_did and not public_did:
        print("No public DID")

    # Debug settings
    test_seed = context.settings.get("debug.seed")
    if context.settings.get("debug.enabled"):
        if not test_seed:
            test_seed = "testseed000000000000000000000001"
    if test_seed:
        await wallet.create_local_did(
            seed=test_seed, metadata={"endpoint": "1.2.3.4:8021"}
        )

    return public_did



import json

from asynctest import TestCase as AsyncTestCase
from marshmallow import EXCLUDE, fields

from ...protocols.didcomm_prefix import DIDCommPrefix
from ...wallet.basic import BasicWallet
from ...wallet.util import bytes_to_b64

from ..agent_message import AgentMessage, AgentMessageSchema
from ..decorators.signature_decorator import SignatureDecorator
from ..decorators.trace_decorator import TraceReport, TRACE_LOG_TARGET
from ..models.base import BaseModelError


class SignedAgentMessage(AgentMessage):
    """Signed agent message tests"""

    class Meta:
        """Meta data"""

        handler_class = None
        schema_class = "SignedAgentMessageSchema"
        message_type = "signed-agent-message"

    def __init__(self, value: str = None, **kwargs):
        super().__init__(**kwargs)
        self.value = value


class SignedAgentMessageSchema(AgentMessageSchema):
    """Utility schema"""

    class Meta:
        model_class = SignedAgentMessage
        signed_fields = ("value",)
        unknown = EXCLUDE

    value = fields.Str(required=True)


class BasicAgentMessage(AgentMessage):
    """Simple agent message implementation"""

    class Meta:
        """Meta data"""

        schema_class = AgentMessageSchema
        message_type = "basic-message"


class TestAgentMessage(AsyncTestCase):
    """Tests agent message."""

    def test_init(self):
        """Tests init class"""

        class BadImplementationClass(AgentMessage):
            """Test utility class."""

        message = SignedAgentMessage()
        message._id = "12345"

        with self.assertRaises(TypeError) as context:
            BadImplementationClass()  # pylint: disable=E0110
        assert "Can't instantiate abstract" in str(context.exception)

        BadImplementationClass.Meta.schema_class = "AgentMessageSchema"
        with self.assertRaises(TypeError) as context:
            BadImplementationClass()  # pylint: disable=E0110
        assert "Can't instantiate abstract" in str(context.exception)

    async def test_field_signature(self):
        wallet = BasicWallet()
        key_info = await wallet.create_signing_key()

        msg = SignedAgentMessage()
        msg.value = None
        with self.assertRaises(BaseModelError) as context:
            await msg.sign_field("value", key_info.verkey, wallet)
        assert "field has no value for signature" in str(context.exception)

        msg.value = "Test value"
        with self.assertRaises(BaseModelError) as context:
            msg.serialize()
        assert "Missing signature for field" in str(context.exception)

        await msg.sign_field("value", key_info.verkey, wallet)
        sig = msg.get_signature("value")
        assert isinstance(sig, SignatureDecorator)

        assert await sig.verify(wallet)
        assert await msg.verify_signed_field("value", wallet) == key_info.verkey
        assert await msg.verify_signatures(wallet)

        with self.assertRaises(BaseModelError) as context:
            await msg.verify_signed_field("value", wallet, "bogus-verkey")
        assert "Signer verkey of signature does not match" in str(context.exception)

        serial = msg.serialize()
        assert "value~sig" in serial and "value" not in serial

        (_, timestamp) = msg._decorators.field("value")["sig"].decode()
        tamper_deco = await SignatureDecorator.create("tamper", key_info.verkey, wallet)
        msg._decorators.field("value")["sig"].sig_data = tamper_deco.sig_data
        with self.assertRaises(BaseModelError) as context:
            await msg.verify_signed_field("value", wallet)
        assert "Field signature verification failed" in str(context.exception)
        assert not await msg.verify_signatures(wallet)

        msg.value = "Test value"
        msg._decorators.field("value").pop("sig")
        with self.assertRaises(BaseModelError) as context:
            await msg.verify_signed_field("value", wallet)
        assert "Missing field signature" in str(context.exception)

        loaded = SignedAgentMessage.deserialize(serial)
        assert isinstance(loaded, SignedAgentMessage)
        assert await loaded.verify_signed_field("value", wallet) == key_info.verkey

    async def test_assign_thread(self):
        msg = BasicAgentMessage()
        assert msg._thread_id == msg._id
        reply = BasicAgentMessage()
        reply.assign_thread_from(msg)
        assert reply._thread_id == msg._thread_id
        assert reply._thread_id != reply._id

        msg.assign_thread_id(None, None)
        assert not msg._thread

    async def test_add_tracing(self):
        msg = BasicAgentMessage()
        msg.add_trace_decorator()
        tracer = msg._trace
        assert tracer.target == TRACE_LOG_TARGET
        assert tracer.full_thread == True

        trace_report = TraceReport(
            msg_id=msg._id,
            thread_id=msg._thread_id,
            traced_type=msg._type,
            timestamp="123456789.123456",
            str_time="2019-01-01 12:34:56.7",
            handler="function.START",
            ellapsed_milli=27,
            outcome="OK! ...",
        )
        msg.add_trace_report(trace_report)
        tracer = msg._trace
        trace_reports = tracer.trace_reports
        assert len(trace_reports) == 1
        msg_trace_report = trace_reports[0]
        assert msg_trace_report.msg_id == msg._id
        assert msg_trace_report.thread_id == msg._thread_id
        assert msg_trace_report.handler == trace_report.handler
        assert msg_trace_report.ellapsed_milli == trace_report.ellapsed_milli
        assert msg_trace_report.traced_type == msg._type
        assert msg_trace_report.outcome == trace_report.outcome

        msg2 = BasicAgentMessage()
        msg2.assign_thread_from(msg)
        msg2.assign_trace_from(msg)
        tracer = msg2._trace
        trace_reports = tracer.trace_reports
        assert len(trace_reports) == 1

        trace_report2 = TraceReport(
            msg_id=msg2._id,
            thread_id=msg2._thread_id,
            traced_type=msg2._type,
            timestamp="123456789.123456",
            str_time="2019-01-01 12:34:56.7",
            handler="function.END",
            ellapsed_milli=72,
            outcome="A OK! ...",
        )
        msg2.add_trace_report(trace_report2)
        tracer = msg2._trace
        trace_reports = tracer.trace_reports
        assert len(trace_reports) == 2
        msg_trace_report = trace_reports[1]
        assert msg_trace_report.msg_id == msg2._id
        assert msg_trace_report.thread_id == msg2._thread_id
        assert msg_trace_report.handler == trace_report2.handler
        assert msg_trace_report.ellapsed_milli == trace_report2.ellapsed_milli
        assert msg_trace_report.traced_type == msg2._type
        assert msg_trace_report.outcome == trace_report2.outcome

        print("tracer:", tracer.serialize())

        msg3 = BasicAgentMessage()
        msg.add_trace_decorator()
        assert msg._trace


class TestAgentMessageSchema(AsyncTestCase):
    """Tests agent message schema."""

    def test_init_x(self):
        """Tests init class"""

        class BadImplementationClass(AgentMessageSchema):
            """Test utility class."""

        with self.assertRaises(TypeError) as context:
            BadImplementationClass()
        assert "Can't instantiate abstract" in str(context.exception)

    def test_extract_decorators_x(self):
        for serial in [
            {
                "@type": "signed-agent-message",
                "@id": "030ac9e6-0d60-49d3-a8c6-e7ce0be8df5a",
                "value": "Test value",
            },
            {
                "@type": "signed-agent-message",
                "@id": "030ac9e6-0d60-49d3-a8c6-e7ce0be8df5a",
                "value": "Test value",
                "value~sig": {
                    "@type": DIDCommPrefix.qualify_current(
                        "signature/1.0/ed25519Sha512_single"
                    ),
                    "signature": (
                        "-OKdiRRQu-xbVGICg1J6KV_6nXLLzYRXr8BZSXzoXimytBl"
                        "O8ULY7Nl1lQPqahc-XQPHiBSVraLM8XN_sCzdCg=="
                    ),
                    "sig_data": "AAAAAF8bIV4iVGVzdCB2YWx1ZSI=",
                    "signer": "7VA3CaF9jaTuRN2SGmekANoja6Js4U51kfRSbpZAfdhy",
                },
            },
            {
                "@type": "signed-agent-message",
                "@id": "030ac9e6-0d60-49d3-a8c6-e7ce0be8df5a",
                "superfluous~sig": {
                    "@type": DIDCommPrefix.qualify_current(
                        "signature/1.0/ed25519Sha512_single"
                    ),
                    "signature": (
                        "-OKdiRRQu-xbVGICg1J6KV_6nXLLzYRXr8BZSXzoXimytBl"
                        "O8ULY7Nl1lQPqahc-XQPHiBSVraLM8XN_sCzdCg=="
                    ),
                    "sig_data": "AAAAAF8bIV4iVGVzdCB2YWx1ZSI=",
                    "signer": "7VA3CaF9jaTuRN2SGmekANoja6Js4U51kfRSbpZAfdhy",
                },
            },
        ]:
            with self.assertRaises(BaseModelError) as context:
                SignedAgentMessage.deserialize(serial)

    def test_serde(self):
        serial = {
            "@type": "signed-agent-message",
            "@id": "030ac9e6-0d60-49d3-a8c6-e7ce0be8df5a",
            "value~sig": {
                "@type": DIDCommPrefix.qualify_current(
                    "signature/1.0/ed25519Sha512_single"
                ),
                "signature": (
                    "-OKdiRRQu-xbVGICg1J6KV_6nXLLzYRXr8BZSXzoXimytBl"
                    "O8ULY7Nl1lQPqahc-XQPHiBSVraLM8XN_sCzdCg=="
                ),
                "sig_data": "AAAAAF8bIV4iVGVzdCB2YWx1ZSI=",
                "signer": "7VA3CaF9jaTuRN2SGmekANoja6Js4U51kfRSbpZAfdhy",
            },
        }
        result = SignedAgentMessage.deserialize(serial)
        result.serialize()



"""Default storage provider classes."""

import logging

from ..config.base import BaseProvider, BaseInjector, BaseSettings
from ..utils.classloader import ClassLoader
from ..wallet.base import BaseWallet

LOGGER = logging.getLogger(__name__)


class StorageProvider(BaseProvider):
    """Provider for the default configurable storage classes."""

    STORAGE_TYPES = {
        "basic": "aries_cloudagent.storage.basic.BasicStorage",
        "indy": "aries_cloudagent.storage.indy.IndyStorage",
        "postgres_storage": "aries_cloudagent.storage.indy.IndyStorage",
    }

    async def provide(self, settings: BaseSettings, injector: BaseInjector):
        """Create and return the storage instance."""
        wallet: BaseWallet = await injector.inject(BaseWallet)

        wallet_type = settings.get_value("wallet.type", default="basic").lower()
        storage_default_type = "indy" if wallet_type == "indy" else "basic"
        storage_type = settings.get_value(
            "storage_type", default=storage_default_type
        ).lower()
        storage_class = self.STORAGE_TYPES.get(storage_type, storage_type)
        storage = ClassLoader.load_class(storage_class)(wallet)
        return storage
